# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Run feature depending on provided args"""
import subprocess
import sys
from pathlib import Path
from typing import List, Optional

from scargo.commands.build import scargo_build
from scargo.config import ScargoTarget, Target
from scargo.config_utils import prepare_config
from scargo.logger import get_logger

logger = get_logger()


def scargo_run(
    bin_path: Optional[Path], profile: str, params: List[str], skip_build: bool
) -> None:
    """
    Run command from CLI

    :param str bin_path: path to bin file
    :param Path profile: build profile name
    :param str params: params for bin file
    :param bool skip_build: skip build step
    :return: None
    """
    logger.info('Running "%s" build', profile)
    config = prepare_config()

    if not config.project.is_x86():
        logger.info(
            "Running non x86 projects on x86 architecture is not implemented yet."
        )
        sys.exit(1)

    if not skip_build:
        scargo_build(profile, ScargoTarget.x86)

    if bin_path:
        bin_file_name = bin_path.name
        bin_file_path = bin_path.parent
        try:
            subprocess.check_call([f"./{bin_file_name}"] + params, cwd=bin_file_path)
        except subprocess.CalledProcessError:
            logger.error(f"Bin file '{bin_path}' not found!")
    else:
        x86_target = Target.get_target_by_id(ScargoTarget.x86.value)
        bin_dir = config.project_root / x86_target.get_profile_build_dir() / "bin"
        if bin_dir.is_dir():
            first_bin = next(bin_dir.iterdir())
            # Run project
            try:
                subprocess.check_call([f"./{first_bin.name}"] + params, cwd=bin_dir)
            except subprocess.CalledProcessError:
                logger.error("Unable to run bin file")
        else:
            logger.error(f"Bin dir '{bin_dir}' not found!")



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import subprocess
import sys
from pathlib import Path
from shutil import copyfile
from typing import Optional

from scargo.config import Config
from scargo.config_utils import prepare_config
from scargo.file_generators.mock_gen import generate_mocks
from scargo.file_generators.ut_gen import generate_ut
from scargo.global_values import SCARGO_PKG_PATH
from scargo.logger import get_logger
from scargo.target_helpers.esp32_helper import (
    OUT_FS_DIR,
    gen_fs_esp32,
    gen_single_binary_esp32,
)

logger = get_logger()


def scargo_gen(
    profile: str,
    gen_ut: Optional[Path],
    gen_mock: Optional[Path],
    certs: Optional[str],
    certs_mode: Optional[str],
    certs_input: Optional[Path],
    certs_passwd: Optional[str],
    fs: bool,
    single_bin: bool,
) -> None:
    config = prepare_config()

    if gen_ut:
        generate_ut(gen_ut, config)

    if gen_mock:
        if gen_mock.suffix not in (".h", ".hpp"):
            logger.error("Not a header file. Please chose .h or .hpp file.")
            sys.exit(1)
        if generate_mocks(gen_mock, config):
            logger.info(f"Generated: {gen_mock}")
        else:
            logger.info(f"Skipping: {gen_mock}")

    if certs:
        generate_certs(
            certs, certs_mode, certs_input, certs_passwd, config.project_root
        )

    if fs:
        generate_fs(config)

    if single_bin:
        gen_single_binary(profile, config)


def generate_certs(
    device_name: str,
    mode_for_certs: Optional[str],
    certs_intermediate_dir: Optional[Path],
    certs_passwd: Optional[str],
    project_path: Path,
) -> None:
    internal_certs_dir = Path(SCARGO_PKG_PATH, "certs")
    projects_builds_path = project_path / "build"
    certs_out_dir = projects_builds_path / "certs"
    if not certs_intermediate_dir:
        certs_intermediate_dir = projects_builds_path / "certs"
    if not certs_passwd:
        certs_passwd = "1234"

    if mode_for_certs == "device":
        mode_for_certs = "Device-certificate"
    else:
        mode_for_certs = "All-certificates"

    certs_out_dir.mkdir(parents=True, exist_ok=True)
    subprocess.call(
        [
            internal_certs_dir / "generateAllCertificates.sh",
            "--name",
            device_name,
            "--mode",
            mode_for_certs,
            "--output",
            certs_out_dir,
            "--input",
            certs_intermediate_dir,
            "--passwd",
            certs_passwd,
        ]
    )

    certs_out_azure_dir = certs_out_dir / "azure"
    certs_out_uc_dir = project_path / OUT_FS_DIR

    certs_out_uc_dir.mkdir(parents=True, exist_ok=True)
    certs_out_azure_dir.mkdir(parents=True, exist_ok=True)

    try:
        copyfile(
            certs_out_dir / "certs" / "iot-device.cert.pem",
            certs_out_uc_dir / "device_cert.pem",
        )
        copyfile(
            certs_out_dir / "private" / "iot-device.key.pem",
            certs_out_uc_dir / "device_priv_key.pem",
        )
        copyfile(certs_out_dir / "ca.pem", certs_out_uc_dir / "ca.pem")

        copyfile(
            certs_out_dir / "certs" / "azure-iot-test-only.root.ca.cert.pem",
            certs_out_azure_dir / f"{device_name}-root-ca.pem",
        )
        logger.info("Cert generated for dev: %s", device_name)
    except FileNotFoundError as e:
        logger.error("Failed to copy certificate %s", e.filename)
        sys.exit(1)


def generate_fs(config: Config) -> None:
    if config.project.is_esp32():
        gen_fs_esp32(config)
    else:
        logger.warning("Gen --fs command not supported for this target yet.")


def gen_single_binary(build_profile: str, config: Config) -> None:
    if config.project.is_esp32():
        gen_single_binary_esp32(build_profile, config)
    else:
        logger.warning("Gen --bin command not supported for this target yet.")



import platform
import subprocess
import sys
from pathlib import Path

from scargo.logger import get_logger

logger = get_logger()


def find_program_path(program_name: str) -> Path:
    cmd = "where" if platform.system() == "Windows" else "/usr/bin/which"
    try:
        cmd_output = subprocess.check_output([cmd, program_name])
        return Path(cmd_output.decode("utf-8").strip())
    except subprocess.CalledProcessError:
        logger.error("%s not installed or not added to PATH", program_name)
        logger.info("Please install %s or add to PATH", program_name)
        sys.exit(1)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import os
import platform
import subprocess
import sys
from pathlib import Path
from typing import List, Optional, Tuple

from scargo.config import ScargoTarget, Target
from scargo.config_utils import prepare_config
from scargo.file_generators.vscode_gen import generate_launch_json
from scargo.logger import get_logger
from scargo.target_helpers import atsam_helper, stm32_helper
from scargo.utils.sys_utils import find_program_path

if platform.system() == "Windows":
    from subprocess import DETACHED_PROCESS  # type: ignore[attr-defined]
else:
    from subprocess import PIPE

logger = get_logger()


class _ScargoFlash:
    FLASH_SUPPORTED_TARGETS = [
        ScargoTarget.stm32,
        ScargoTarget.esp32,
        ScargoTarget.atsam,
    ]

    def __init__(
        self,
        flash_profile: str,
        port: Optional[str],
        target: Optional[ScargoTarget],
        app: bool,
        file_system: bool,
        erase_memory: bool,
        bank: Optional[int],
    ):
        self._flash_profile = flash_profile
        self._port = port
        self._app = app
        self._file_system = file_system
        self._erase_memory = erase_memory
        self._bank = bank

        self._config = prepare_config()
        self._target = self._initialize_target(target)
        self._validate_target()
        self._validate_erase_memory()

    def _initialize_target(self, target: Optional[ScargoTarget]) -> Target:
        if target:
            if target.value not in self._config.project.target_id:
                logger.error(f"Target {target.value} not defined in scargo toml")
                sys.exit(1)
            return Target.get_target_by_id(target.value)
        return self._get_first_supported_target()

    def _validate_target(self) -> None:
        if self._target.id not in self.FLASH_SUPPORTED_TARGETS:
            logger.error("Flash command not supported for this target yet.")
            sys.exit(1)

    def _validate_erase_memory(self) -> None:
        if self._erase_memory and self._target.id != ScargoTarget.stm32:
            logger.info(f"--no-erase has not effect on {self._target.id}")

    def _get_first_supported_target(self) -> Target:
        for target in self._config.project.target:
            if target.id in self.FLASH_SUPPORTED_TARGETS:
                return target
        logger.error("Project does not contain target that supports flashing")
        sys.exit(1)

    def _get_binary_and_elf_paths(self) -> Tuple[Path, Path]:
        project_root = self._config.project_root
        elf_path = project_root / self._target.get_bin_path(
            self._config.project.name.lower(), self._flash_profile
        )
        bin_path = elf_path.with_suffix(".bin")
        return bin_path, elf_path

    def _check_bin_path(self, bin_path: Path) -> None:
        if not bin_path.is_file():
            logger.error("%s does not exist", bin_path)
            logger.info(
                "Did you run scargo build --profile %s --target %s",
                self._flash_profile,
                self._target.id,
            )
            sys.exit(1)

    def flash_target(self) -> None:
        if self._target.id == ScargoTarget.atsam:
            self._flash_atsam()
        elif self._target.id == ScargoTarget.esp32:
            self._flash_esp32()
        elif self._target.id == ScargoTarget.stm32:
            self._flash_stm32()

    def _flash_atsam(self) -> None:
        openocd_path = find_program_path("openocd")
        gdb_multiarch_path = find_program_path("gdb-multiarch")

        bin_path, elf_path = self._get_binary_and_elf_paths()
        self._check_bin_path(elf_path)
        self._check_bin_path(bin_path)

        atsam_helper.generate_gdb_script(Path(".devcontainer"), self._config, bin_path)
        generate_launch_json(Path(".vscode"), self._config, elf_path)

        openocd_process = None
        try:
            openocd_process = self._start_openocd(openocd_path)

            gdb_command = [
                str(gdb_multiarch_path),
                f"{elf_path}",
                "--command=.devcontainer/atsam-gdb.script",
                "--batch",
            ]
            subprocess.run(gdb_command, cwd=self._config.project_root, check=True)
        except subprocess.CalledProcessError as e:
            logger.error("Failed to flash atsam target: %s", e.stderr)
            sys.exit(1)
        finally:
            self._cleanup_openocd(openocd_process)

    def _flash_esp32(self) -> None:
        try:
            if self._app:
                self._flash_esp32_app()
            elif self._file_system:
                self._flash_esp32_fs()
            else:
                self._flash_esp32_default()

            if self._bank is not None:
                self._switch_bank_esp32()

        except subprocess.CalledProcessError as e:
            logger.error("Failed to flash esp32 target: %s", e.stderr)
            sys.exit(1)

    def _flash_stm32(self) -> None:
        bin_path, elf_path = self._get_binary_and_elf_paths()
        self._check_bin_path(elf_path)
        self._check_bin_path(bin_path)
        flash_start = hex(self._config.get_stm32_config().flash_start)

        if self._erase_memory:
            command = ["sudo", "st-flash"]
            if self._port:
                command.append(f"--serial={self._port}")
            command.append("erase")
            subprocess.run(command, check=True)

        command = ["sudo", "st-flash", "--reset"]
        if self._port:
            command.append(f"--serial={self._port}")
        command.extend(["write", str(bin_path), flash_start])
        subprocess.run(command, check=True)

        stm32_helper.generate_openocd_script(Path(".devcontainer"), self._config)
        generate_launch_json(Path(".vscode"), self._config, elf_path)

    def _start_openocd(self, openocd_path: Path) -> subprocess.Popen:  # type: ignore[type-arg]
        openocd_script = self._config.project_root / ".devcontainer/openocd-script.cfg"
        if platform.system() == "Windows":
            return subprocess.Popen(
                [str(openocd_path), "-f", str(openocd_script)],
                creationflags=DETACHED_PROCESS,
            )
        return subprocess.Popen(
            ["sudo", str(openocd_path), "-f", str(openocd_script)],
            stdin=PIPE,
            stdout=PIPE,
            stderr=PIPE,
        )

    def _cleanup_openocd(self, openocd_process: Optional[subprocess.Popen]) -> None:  # type: ignore[type-arg]
        if openocd_process is not None:
            if platform.system() == "Windows":
                openocd_process.terminate()
            else:
                os.system(f"sudo pkill -9 -P {openocd_process.pid}")

    def _flash_esp32_app(self) -> None:
        bin_dir_path = self._target.get_bin_dir_path(self._flash_profile)
        bin_path = Path(bin_dir_path, f"{self._config.project.name}.bin")
        command = self._build_esp32_flash_command(
            "parttool.py", "ota_0", [f"--input={bin_path}"]
        )
        subprocess.run(command, cwd=self._config.project_root, check=True)

    def _flash_esp32_fs(self) -> None:
        command = self._build_esp32_flash_command(
            "parttool.py", "spiffs", ["--input=build/spiffs.bin"]
        )
        subprocess.run(command, cwd=self._config.project_root, check=True)

    def _flash_esp32_default(self) -> None:
        command = self._build_esp32_flash_command(
            "esptool.py", None, ["write_flash", "@flash_args"]
        )
        bin_dir = self._config.project_root / self._target.get_bin_dir_path(
            self._flash_profile
        )
        subprocess.run(command, cwd=bin_dir, check=True)

    def _build_esp32_flash_command(
        self, tool: str, partition_name: Optional[str], extra_args: List[str]
    ) -> List[str]:
        command = [tool]
        if self._port:
            command.append(f"--port={self._port}")

        if tool == "esptool.py":
            chip = self._config.get_esp32_config().chip
            command.extend(["--chip", chip])

        if partition_name:
            command.extend(["write_partition", f"--partition-name={partition_name}"])
        command.extend(extra_args)
        return command

    def _switch_bank_esp32(self) -> None:
        command = ["otatool.py", "switch_ota_partition", "--slot", str(self._bank)]

        try:
            logger.info("Bank switching to: [%d]", self._bank)
            subprocess.run(command, cwd=self._config.project_root, check=True)
        except subprocess.CalledProcessError as e:
            logger.error("Command failed with return code %s", str(e.returncode))
        except Exception as e:  # pylint: disable=broad-except
            logger.error("An error occurred: %s", str(e))


def scargo_flash(
    flash_profile: str,
    port: Optional[str],
    target: Optional[ScargoTarget],
    app: bool,
    file_system: bool,
    erase_memory: bool,
    bank: Optional[int] = None,
) -> None:
    flasher = _ScargoFlash(
        flash_profile, port, target, app, file_system, erase_memory, bank
    )
    flasher.flash_target()



import subprocess
from pathlib import Path
from typing import Dict, List

from scargo.config import Config
from scargo.logger import get_logger

logger = get_logger()


DEFAULT_PROFILES = ["Debug", "Release", "RelWithDebInfo", "MinSizeRel"]


def conan_add_remote(project_path: Path, config: Config) -> None:
    """
    Add conan remote repository

    :param Path project_path: path to project
    :param Config config:
    :return: None
    """
    remotes_without_user = _get_remotes_without_user(config.conan.repo)
    for repo_name, repo_url in config.conan.repo.items():
        try:
            subprocess.run(
                ["conan", "remote", "add", repo_name, repo_url],
                cwd=project_path,
                check=True,
                stderr=subprocess.PIPE,
            )
        except subprocess.CalledProcessError as e:
            if b"already exists in remotes" not in e.stderr:
                logger.error(e.stderr.decode().strip())
                logger.error("Unable to add remote repository")
            else:
                pass
        if repo_name in remotes_without_user:
            conan_remote_login(repo_name)


def conan_remote_login(remote: str) -> None:
    """
    Add conan user

    :param str remote: name of remote repository
    :return: None
    """
    remote_login_command = ["conan", "remote", "login", remote]
    logger.info("Login to conan remote %s", remote)

    try:
        subprocess.run(remote_login_command, check=True)
    except subprocess.CalledProcessError:
        logger.error(f"Unable to log in to conan remote {remote}")


def conan_source(project_dir: Path) -> None:
    try:
        subprocess.run(["conan", "source", "."], cwd=project_dir, check=True)
    except subprocess.CalledProcessError:
        logger.error("Unable to source")


def _get_remotes_without_user(conan_remotes: Dict[str, str]) -> List[str]:
    result = subprocess.run(
        ["conan", "remote", "list-users"], check=True, stdout=subprocess.PIPE
    )
    user_list_stdout = result.stdout.decode().splitlines()

    no_users = []
    for index, line in enumerate(user_list_stdout):
        if remote_line := line.strip().split(":")[0]:
            if (
                remote_line in conan_remotes
                and "No user" in user_list_stdout[index + 1]
            ):
                no_users.append(remote_line)

    return no_users



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from scargo.config import Config, Stm32Config
from scargo.file_generators.base_gen import write_template

script_dir = Path(__file__).parent
atmel_arxml_path = script_dir / "atmel.xml"

ADDITIONAL_CPU_DATA = {"cortex-m23": ["atsaml10e16a"]}


@dataclass
class STM32Scrips:
    openocd_cfg = "openocd-script.cfg"


def create_stm32_config(chip: Optional[str]) -> Optional[Stm32Config]:
    if chip is None:
        return None
    return Stm32Config(chip=chip)


def generate_openocd_script(outdir: Path, config: Config) -> None:
    chip = config.get_stm32_config().chip

    chip_script = f"target/{chip[:7].lower()}x.cfg"
    if not Path("/usr/share/openocd/scripts", chip_script).exists():
        chip_script = f"target/{chip[:7].lower()}.cfg"

    write_template(
        outdir / STM32Scrips.openocd_cfg,
        "docker/stm32-openocd.cfg.j2",
        {"chip_script": chip_script},
    )



import re
from pathlib import Path
from typing import Iterable, List, Tuple


def finditer_with_line_numbers(
    pattern: re.Pattern, string: str, flags: int = 0  # type: ignore[type-arg]
) -> Iterable[Tuple[re.Match, int]]:  # type: ignore[type-arg]
    """
    Version of re.finditer that additionally returns the line number for each match.

    :param pattern: A compiled regular expression pattern.
    :param string: The string to be searched.
    :param flags: Standard re flags to modify the regex behavior
    :yield: Tuples of (match object, line number), where line number is the line on which the match starts.
    """
    matches = list(re.finditer(pattern, string, flags))
    if not matches:
        return

    end = matches[-1].start()
    newline_table = {-1: 0}
    for i, m in enumerate(re.finditer("\\n", string), 1):
        offset = m.start()
        if offset > end:
            break
        newline_table[offset] = i

    for m in matches:
        newline_offset = string.rfind("\n", 0, m.start())
        line_number = newline_table[newline_offset]
        yield (m, line_number)


def extract_block_comments(content: str) -> List[str]:
    """
    Extracts block comments from the given content.

    :param content: The string to extract comments from.
    :return: A list of block comments.
    """
    pattern = re.compile(r"/\*[\s\S]*?\*/", re.DOTALL)
    return pattern.findall(content)


def extract_grouped_line_comments(content: str) -> List[str]:
    """
    Extracts and groups line comments from the given content.

    :param content: The string to extract comments from.
    :param pattern: The compiled regex pattern for line comments.
    :return: A list of grouped line comments.
    """
    pattern = re.compile(r"//.*", re.MULTILINE)
    grouped_comments = []
    current_section: List[Tuple[str, int]] = []

    for comment, line_number in finditer_with_line_numbers(pattern, content):
        if current_section and line_number > current_section[-1][1] + 1:
            grouped_comments.append(
                "\n".join(comment[0] for comment in current_section)
            )
            current_section = []
        current_section.append((comment.group(), line_number))

    if current_section:
        grouped_comments.append("\n".join(comment[0] for comment in current_section))

    return grouped_comments


def extract_comment_sections(file_path: Path) -> List[str]:
    """
    Extracts comment sections from the given file.

    :param file_path: A Path object pointing to the file.
    :return: A list of comment sections.
    """
    content = file_path.read_text()

    block_comments = extract_block_comments(content)
    grouped_line_comments = extract_grouped_line_comments(content)

    return block_comments + grouped_line_comments



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import toml
from pydantic import BaseModel, Extra, Field, root_validator

from scargo.global_values import SCARGO_DEFAULT_BUILD_ENV, SCARGO_DOCKER_ENV


class ScargoTarget(Enum):
    atsam = "atsam"
    esp32 = "esp32"
    stm32 = "stm32"
    x86 = "x86"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, str):
            return self.value == other
        if isinstance(other, ScargoTarget):
            return self.value == other.value
        return False

    def __hash__(self) -> int:
        return super().__hash__()


CHIP_DEFAULTS = {
    "x86": "",
    "esp32": "esp32",
    "atsam": "ATSAML10E16A",
    "stm32": "STM32L496AG",
}

ESP32_DEFAULT_PARTITIONS = [
    "nvs,      data, nvs,     0x9000,  0x4000,",
    "otadata,  data, ota,     0xd000,  0x2000,",
    "phy_init, data, phy,     0xf000,  0x1000,",
    "ota_0,    app,  ota_0,   0x10000, 0x180000,",
    "ota_1,    app,  ota_1,   0x190000,0x180000,",
    "spiffs,   data, spiffs,  0x310000,0x6000,",
]


class ConfigError(Exception):
    pass


class Config(BaseModel):
    project: "ProjectConfig"
    profiles: Dict[str, "ProfileConfig"] = Field(..., alias="profile")
    check: "ChecksConfig"
    fix: "FixesConfig" = Field(
        default_factory=lambda: FixesConfig()  # pylint: disable=unnecessary-lambda
    )
    doc: "DocConfig" = Field(
        default_factory=lambda: DocConfig()  # pylint: disable=unnecessary-lambda
    )
    tests: "TestConfig"
    dependencies: "Dependencies"
    conan: "ConanConfig"
    atsam: Optional["ATSAMConfig"]
    stm32: Optional["Stm32Config"]
    esp32: Optional["Esp32Config"]
    scargo: "ScargoConfig" = Field(
        default_factory=lambda: ScargoConfig()  # pylint: disable=unnecessary-lambda
    )
    docker_compose: "DockerComposeConfig" = Field(
        default_factory=lambda: DockerComposeConfig(),  # pylint: disable=unnecessary-lambda
        alias="docker-compose",
    )
    project_root: Path

    @property
    def source_dir_path(self) -> Path:
        if self.project.is_esp32():
            # Backwards compatibility for main dir
            main_dir = self.project_root / COMPATIBILITY_ESP32_SRC_DIR
            if main_dir.is_dir():
                return main_dir
        return self.project_root / DEFAULT_SRC_DIR

    @property
    def include_dir_path(self) -> Path:
        return self.source_dir_path / DEFAULT_INCLUDE_DIR

    def get_atsam_config(self) -> "ATSAMConfig":
        if not self.atsam:
            raise ConfigError("No [atsam] section in config")
        return self.atsam

    def get_stm32_config(self) -> "Stm32Config":
        if not self.stm32:
            raise ConfigError("No [stm32] section in config")
        return self.stm32

    def get_esp32_config(self) -> "Esp32Config":
        if not self.esp32:
            raise ConfigError("No [esp32] section in config")
        return self.esp32

    @root_validator
    def validate_special_configs(  # pylint: disable=no-self-argument
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        if "project" in values:
            target_id = values["project"].target_id
            targets_for_validation = [
                ScargoTarget.stm32.value,
                ScargoTarget.esp32.value,
                ScargoTarget.atsam.value,
            ]
            for target in targets_for_validation:
                if target in target_id and not values.get(target):
                    raise ConfigError(f"No [{target}] section in config")

        # Set default value of cmake_build_type - Debug for non-stanard profiles,
        # If profile is on standard_profiles list, use it's name instead
        standard_profiles: List[str] = [
            "Debug",
            "Release",
            "RelWithDebInfo",
            "MinSizeRel",
        ]
        if "profiles" in values:
            for name, profile in values["profiles"].items():
                if profile.cmake_build_type is None:
                    if name in standard_profiles:
                        profile.cmake_build_type = name
                    else:
                        profile.cmake_build_type = "Debug"
        return values


class ProjectConfig(BaseModel):
    name: str
    version: str
    description: Optional[str]
    homepage_url: Optional[str] = Field(default=None, alias="homepage-url")

    bin_name: Optional[str]
    lib_name: Optional[str]
    target_id: Union[str, List[str]] = Field(..., alias="target")
    build_env: str = Field(SCARGO_DEFAULT_BUILD_ENV, alias="build-env")
    docker_file: Path = Field(..., alias="docker-file")
    docker_image_tag: str = Field(..., alias="docker-image-tag")
    in_repo_conan_cache: bool = Field(..., alias="in-repo-conan-cache")

    cc: Optional[str] = None
    cxx: Optional[str] = None
    cxxstandard: str

    cflags: str
    cxxflags: str

    max_build_jobs: Optional[int] = Field(default=None, alias="max-build-jobs")
    cmake_variables: Dict[str, str] = Field(default={}, alias="cmake-variables")

    @property
    def target(self) -> List["Target"]:
        if isinstance(self.target_id, str):
            return Target.get_targets_by_id([self.target_id])
        return Target.get_targets_by_id(self.target_id)

    @property
    def default_target(self) -> "Target":
        if isinstance(self.target_id, str):
            return Target.get_target_by_id(self.target_id)
        return Target.get_target_by_id(self.target_id[0])

    def is_docker_buildenv(self) -> bool:
        return self.build_env == SCARGO_DOCKER_ENV

    def is_x86(self) -> bool:
        return "x86" in self.target_id

    def is_stm32(self) -> bool:
        return "stm32" in self.target_id

    def is_esp32(self) -> bool:
        return "esp32" in self.target_id

    def is_atsam(self) -> bool:
        return "atsam" in self.target_id

    def is_multitarget(self) -> bool:
        return isinstance(self.target_id, list) and len(self.target_id) > 1


class Target(BaseModel):
    id: str
    elf_file_extension: str
    cc: Optional[str] = None
    cxx: Optional[str] = None

    def get_profile_build_dir(self, profile: str = "Debug") -> str:
        return f"build/{self.id}/{profile}"

    def get_conan_profile_name(self, profile: str = "Debug") -> str:
        return f"{self.id}_{profile}"

    def get_bin_dir_path(self, profile: str = "Debug") -> str:
        build_dir = self.get_profile_build_dir(profile)
        if self.id == ScargoTarget.esp32:
            return build_dir
        return f"{build_dir}/bin"

    def get_bin_path(self, bin_name: str, profile: str = "Debug") -> str:
        bin_dir = self.get_bin_dir_path(profile)
        return f"{bin_dir}/{bin_name}{self.elf_file_extension}"

    @classmethod
    def get_target_by_id(cls, target_id: str) -> "Target":
        return TARGETS[target_id]

    @classmethod
    def get_targets_by_id(cls, target_ids: List[str]) -> List["Target"]:
        return [TARGETS[id] for id in target_ids]


DEFAULT_SRC_DIR = "src"
COMPATIBILITY_ESP32_SRC_DIR = "main"
DEFAULT_INCLUDE_DIR = "include"
TARGETS = {
    ScargoTarget.x86.value: Target(
        id=ScargoTarget.x86.value,
        elf_file_extension="",
        cc="gcc",
        cxx="g++",
    ),
    ScargoTarget.stm32.value: Target(
        id=ScargoTarget.stm32.value, elf_file_extension=".elf"
    ),
    ScargoTarget.esp32.value: Target(
        id=ScargoTarget.esp32.value, elf_file_extension=".elf"
    ),
    ScargoTarget.atsam.value: Target(
        id=ScargoTarget.atsam.value, elf_file_extension=""
    ),
}


class ProfileConfig(BaseModel, extra=Extra.allow):
    cflags: Optional[str]
    cxxflags: Optional[str]
    cc: Optional[str] = None
    cxx: Optional[str] = None
    cmake_build_type: Optional[str] = Field(default=None, alias="cmake-build-type")

    @property
    def extras(self) -> Dict[str, Any]:
        return {
            key: value
            for key, value in dict(self).items()
            if key not in self.__fields__
        }


class ChecksConfig(BaseModel):
    exclude: List[str]
    pragma: "CheckConfig"
    copyright: "CheckConfig"
    todo: "TodoCheckConfig"
    clang_format: "CheckConfig" = Field(..., alias="clang-format")
    clang_tidy: "CheckConfig" = Field(..., alias="clang-tidy")
    cyclomatic: "CheckConfig"


class FixesConfig(BaseModel):
    copyright: "FixConfig" = Field(
        default_factory=lambda: FixConfig()  # pylint: disable=unnecessary-lambda
    )


class CheckConfig(BaseModel):
    description: Optional[str] = None
    exclude: List[str] = Field(default_factory=list)


class FixConfig(BaseModel):
    description: Optional[str] = None


class TodoCheckConfig(CheckConfig):
    keywords: List[str] = Field(default_factory=list)


class DocConfig(BaseModel):
    exclude: List[str] = Field(default_factory=list)


class TestConfig(BaseModel, extra=Extra.allow):
    cc: str
    cxx: str

    cflags: str
    cxxflags: str

    gcov_executable: str = Field(..., alias="gcov-executable")

    @property
    def extras(self) -> Dict[str, Any]:
        return {
            key: value
            for key, value in dict(self).items()
            if key not in self.__fields__
        }


class Dependencies(BaseModel):
    general: List[str] = Field(default_factory=list)
    build: List[str] = Field(default_factory=list)
    tool: List[str] = Field(default_factory=list)
    test: List[str] = Field(default_factory=list)


class ConanConfig(BaseModel):
    repo: Dict[str, str]  # name to url mapping


class Stm32Config(BaseModel):
    chip: str = Field(default=CHIP_DEFAULTS.get("stm32"))
    flash_start: int = Field(default=0x08000000, alias="flash-start")


class ATSAMConfig(BaseModel):
    chip: str = Field(default=CHIP_DEFAULTS.get("atsam"))
    cpu: str = Field()

    @property
    def chip_series(self) -> str:
        return self.chip[2:8].upper()


class Esp32Config(BaseModel):
    chip: str = Field(default=CHIP_DEFAULTS.get("esp32"))
    extra_component_dirs: List[Path] = Field(default=["src"])
    partitions: List[str] = Field(default=ESP32_DEFAULT_PARTITIONS)


class ScargoConfig(BaseModel):
    console_log_level: str = Field(default="INFO", alias="console-log-level")
    file_log_level: str = Field(default="WARNING", alias="file-log-level")
    update_exclude: List[str] = Field(alias="update-exclude", default_factory=list)
    version: Optional[str] = None


class DockerComposeConfig(BaseModel):
    ports: List[str] = Field(default_factory=list)


Config.update_forward_refs()
ProjectConfig.update_forward_refs()
ChecksConfig.update_forward_refs()
FixesConfig.update_forward_refs()
FixConfig.update_forward_refs()
Stm32Config.update_forward_refs()
Esp32Config.update_forward_refs()


def parse_config(config_file_path: Path) -> Config:
    config_obj = toml.load(config_file_path)
    config_obj["project_root"] = config_file_path.parent.absolute()
    config: Config = Config.parse_obj(config_obj)
    return config



# #
# @copyright Copyright (C) 2024 SpyroSoft Solutions S.A. All rights reserved.
# #

import atexit
import os
import signal
import sys
import threading
from types import FrameType
from typing import Optional

import serial  # type: ignore

from scargo.logger import get_logger

logger = get_logger()

AT_EOF = "\r\n"


class CmdLoop:
    def __init__(self, ser: serial.Serial) -> None:  # type: ignore[no-any-unimported]
        """Init of the command loop

        Args:
            ser (serial.Serial): serial instance
        """
        self.ser = ser
        self.running = True

    def run(self) -> None:
        """Infinit keyboard input loop"""
        while self.running:
            cmd = input()
            if not self.parse_cmd(cmd):
                break

    def parse_cmd(self, cmd: str) -> bool:
        """Parse command to the device


        Args:
            cmd (str): command

        Returns:
            bool: send successfully to the device
        """
        try:
            self.ser.write(str.encode(cmd + AT_EOF))
        except Exception as e:  # pylint: disable=broad-except
            logger.error("Unexpected error: %s", e)
            return False

        return True

    def stop(self) -> None:
        """Stop the infinit loop"""
        self.running = False


class SerialReadThread(threading.Thread):
    def __init__(self, ser: serial.Serial, read_block_size: int = 1000) -> None:  # type: ignore[no-any-unimported]
        """Init the serial data

        Args:
            ser (serial.Serial): _description_
            read_block_size (int, optional): _description_. Defaults to 1000.

        Returns:
            Any: _description_
        """
        super().__init__()
        self.ser = ser
        self.read_block_size = read_block_size
        self.running = True

    def run(self) -> None:
        """Infinite serial read/write loop"""
        while self.running:
            if self.ser.closed:
                break
            self.read_serial()

    def read_serial(self) -> None:
        """Read data from the serial interface"""
        if not self.running or self.ser.closed:
            return
        try:
            s = self.ser.read(self.read_block_size)
            if s:
                print(s.decode())
        except Exception as e:  # pylint: disable=broad-except
            if "Attempting to use a port that is not open" in str(e):
                logger.error("Unexpected error: %s", e)
                os._exit(1)
            else:
                logger.error("Unexpected error: %s", e)

    def stop(self) -> None:
        """Stop the infinit loop"""
        self.running = False


class _ScargoMonitor:
    DEFAULT_BAUDRATE = 115200

    def __init__(
        self,
        port: Optional[str],
        baudrate: Optional[int],
    ) -> None:
        self._port = port
        self._baudrate = baudrate or self.DEFAULT_BAUDRATE
        self.read_thread: Optional[SerialReadThread] = None
        self.cmd_loop: Optional[CmdLoop] = None
        self.ser = None

    def serial_monitor(self) -> None:
        self.ser = serial.Serial(self._port, self._baudrate, timeout=2)
        self.read_thread = SerialReadThread(self.ser)
        if self.read_thread is not None:
            self.read_thread.start()

        self.cmd_loop = CmdLoop(self.ser)
        self.setup_signal_handler()
        if self.cmd_loop is not None:
            self.cmd_loop.run()

        if self.read_thread is not None:
            self.read_thread.join()

    def setup_signal_handler(self) -> None:
        def cleanup() -> None:
            if self.cmd_loop:
                self.cmd_loop.stop()
            if self.read_thread and self.read_thread.running:
                self.read_thread.stop()
            if self.ser and not self.ser.closed:
                self.ser.close()

        def signal_handler(_sig: int, _frame: Optional[FrameType]) -> None:
            """handle CTRL+C

            Args:
                _sig (int): Signal number.
                _frame (FrameType): Current stack frame.
            """
            cleanup()
            sys.exit(0)

        atexit.register(cleanup)
        signal.signal(signal.SIGINT, signal_handler)
        sys.stdin.flush()


def scargo_monitor(
    port: Optional[str],
    baudrate: Optional[int],
) -> None:
    monitor = _ScargoMonitor(port, baudrate)
    monitor.serial_monitor()



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #


"""Publish conan package into repository"""
import subprocess
import sys
from pathlib import Path

from scargo.config_utils import prepare_config
from scargo.logger import get_logger
from scargo.utils.conan_utils import conan_add_remote, conan_source

logger = get_logger()


def scargo_publish(repo: str, profile: str = "Release") -> None:
    """
    Publish conan package

    :param str repo: repository name
    :return: None
    """
    config = prepare_config()
    project_path = config.project_root
    project_config = config.project
    project_name = project_config.name

    # For now upload only for default target, in future we can add support for
    # multiple targets and how to handle them
    target = project_config.default_target
    build_dir = Path(project_path, target.get_profile_build_dir(profile))

    if not build_dir.exists():
        logger.error("Build folder for specified build type does not exist")
        logger.info(f"Did you run 'scargo build --profile {profile}'?")
        sys.exit(1)

    conan_add_remote(project_path, config)
    conan_source(project_path)

    profile_name = target.get_conan_profile_name(profile)
    profile_path = f"./config/conan/profiles/{profile_name}"

    # Export package
    try:
        subprocess.run(
            [
                "conan",
                "export-pkg",
                ".",
                "-pr",
                profile_path,
                "-of",
                build_dir,
            ],
            check=True,
            cwd=project_path,
        )
    except subprocess.CalledProcessError:
        logger.error("Unable to export package")
        sys.exit(1)

    # Test if package has been exported successfully
    try:
        subprocess.run(
            [
                "conan",
                "test",
                "test_package",
                f"{project_name}/{config.project.version}",
                "-pr",
                profile_path,
            ],
            check=True,
            cwd=project_path,
        )
    except subprocess.CalledProcessError:
        logger.error("Package test failed")
        sys.exit(1)

    # Upload package to conan remote
    conan_repo = ["-r", repo] if repo else []
    try:
        subprocess.run(
            [
                "conan",
                "upload",
                f"{project_name}",
                *conan_repo,
                "--confirm",
            ],
            check=True,
            cwd=project_path,
        )
    except subprocess.CalledProcessError:
        logger.error("Unable to publish package")
        sys.exit(1)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #


from pathlib import Path
from typing import Any, Dict

from scargo.config import Config
from scargo.file_generators.base_gen import create_file_from_template


class _VSCodeTemplate:
    """
    This class is a container for docker compose yaml files creation with multilayer approach
    """

    def __init__(self, config: Config, vscode_path: Path):
        self.vscode_path = vscode_path
        self._config = config

    def generate_vscode_env(self) -> None:
        """Generate dirs and files"""

        self._create_file_from_template(
            "vscode/tasks.json.j2",
            "tasks.json",
            template_params={"project": self._config.project},
        )

    def generate_launch_json(self, elf_path: Path) -> None:
        self._create_file_from_template(
            "vscode/launch.json.j2",
            "launch.json",
            template_params={"bin_path": elf_path},
        )

    def _create_file_from_template(
        self,
        template_path: str,
        output_filename: str,
        template_params: Dict[str, Any],
        overwrite: bool = True,
    ) -> None:
        create_file_from_template(
            template_path,
            self.vscode_path / output_filename,
            template_params=template_params,
            config=self._config,
            overwrite=overwrite,
        )


def generate_vscode(vscode_path: Path, config: Config) -> None:
    vscode_template = _VSCodeTemplate(config, vscode_path)
    vscode_template.generate_vscode_env()


def generate_launch_json(vscode_path: Path, config: Config, elf_path: Path) -> None:
    vscode_template = _VSCodeTemplate(config, vscode_path)
    vscode_template.generate_launch_json(elf_path)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import subprocess

from scargo.config import Config
from scargo.file_generators.base_gen import create_file_from_template


def generate_conanfile(config: Config) -> None:
    create_file_from_template(
        "conan/conanfile.py.j2",
        "conanfile.py",
        template_params={"config": config},
        config=config,
    )
    create_file_from_template(
        "conan/conanfiletest.j2",
        "tests/conanfile.py",
        template_params={"config": config},
        config=config,
    )


def generate_conanprofile(config: Config) -> None:
    profiles = config.profiles.keys()

    if config.project.is_stm32():
        create_file_from_template(
            "conan/toolchain/stm32_gcc_toolchain.cmake.j2",
            "config/conan/profiles/stm32_gcc_toolchain.cmake",
            template_params={"config": config},
            config=config,
        )

    if config.project.is_atsam():
        create_file_from_template(
            "conan/toolchain/arm_gcc_toolchain.cmake.j2",
            "config/conan/profiles/arm_gcc_toolchain.cmake",
            template_params={"config": config},
            config=config,
        )

    for target in config.project.target:
        for profile in profiles:
            profile_name = target.get_conan_profile_name(profile)
            create_file_from_template(
                f"conan/profile_{target.id}.j2",
                f"config/conan/profiles/{profile_name}",
                template_params={
                    "config": config,
                    "profile": profile,
                },
                config=config,
            )


def conan_add_default_profile_if_missing() -> None:
    result = subprocess.run(
        ["conan", "profile", "list"],
        stdout=subprocess.PIPE,
        check=True,
    )
    if b"default" not in result.stdout.splitlines():
        subprocess.run(
            ["conan", "profile", "detect"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )



import os
import re
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Optional

from scargo.config import Config, Esp32Config
from scargo.logger import get_logger

logger = get_logger()
OUT_FS_DIR = Path("build", "fs")


def create_esp32_config(chip: Optional[str]) -> Optional[Esp32Config]:
    if chip is None:
        return None
    return Esp32Config(chip=chip)


def gen_fs_esp32(config: Config) -> None:
    command = []
    partition_list = config.get_esp32_config().partitions
    fs_size = 0
    for i in partition_list:
        split_list = i.split(",")
        if "spiffs" in split_list[0]:
            fs_size = int(re.sub(",", "", split_list[4]), 16)

    try:
        project_path = config.project_root
        fs_in_dir = project_path / "main/fs"
        fs_out_dir = project_path / OUT_FS_DIR
        fs_out_bin = project_path / "build/spiffs.bin"
        fs_in_dir.mkdir(parents=True, exist_ok=True)
        fs_out_dir.mkdir(parents=True, exist_ok=True)

        shutil.copytree(fs_in_dir, fs_out_dir, dirs_exist_ok=True)

        idf_path = os.environ.get("IDF_PATH")
        command = [
            f"{idf_path}/components/spiffs/spiffsgen.py",
            str(fs_size),
            str(fs_out_dir),
            str(fs_out_bin),
        ]

        subprocess.run(command, cwd=project_path, check=True)
        logger.info("Generated %s of size:%s", fs_out_bin, fs_size)

    except subprocess.CalledProcessError:
        logger.error("%s fail", command)
        sys.exit(1)


def gen_single_binary_esp32(build_profile: str, config: Config) -> None:
    partition_list = config.get_esp32_config().partitions
    chip = config.get_esp32_config().chip

    build_dir = config.project_root / "build" / build_profile
    flasher_args_path = build_dir / "flash_args"
    if not flasher_args_path.is_file():
        logger.warning("%s does not exists", flasher_args_path)
        logger.info("Did you run scargo build --profile %s", build_profile)
        sys.exit(1)

    line_list = flasher_args_path.read_text().split()

    spiffs_addr = None
    for i in partition_list:
        if "spiffs" in i:
            split_list = i.split(",")
            spiffs_addr = split_list[3].strip()
            break

    flash_size = "4MB"
    for index, arg in enumerate(line_list):
        if "flash_size" in arg:
            flash_size = line_list[index + 1]
        if arg.endswith(".bin"):
            line_list[index] = str(build_dir / arg)

    command = [
        "esptool.py",
        "--chip",
        chip,
        "merge_bin",
        "-o",
        "build/flash_image.bin",
        "--fill-flash-size",
        f"{flash_size}",
    ]
    command.extend(line_list)

    if spiffs_addr:
        command.extend([spiffs_addr, "build/spiffs.bin"])

    try:
        logger.info("Running: %s", " ".join(command))
        subprocess.run(command, cwd=config.project_root, check=True)
    except subprocess.CalledProcessError:
        logger.error("Generation of single binary failed")
        sys.exit(1)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #


import os
import platform
import subprocess
import sys
from pathlib import Path
from time import sleep
from typing import List, Optional

from scargo.config import CHIP_DEFAULTS, Config, ScargoTarget
from scargo.config_utils import get_target_or_default, prepare_config
from scargo.logger import get_logger
from scargo.utils.sys_utils import find_program_path

logger = get_logger()


EMEDDED_GDB_SETTINGS = "--eval-command=target extended-remote localhost:3333"


class _ScargoDebug:
    SUPPORTED_TARGETS = ["x86", "stm32", "esp32", "atsam"]

    def __init__(
        self, config: Config, bin_path: Optional[Path], target: Optional[ScargoTarget]
    ):
        self._config = config
        self._target = get_target_or_default(config, target)

        logger.info(f"Running scargo debug for {self._target.id} target")
        if self._target.id not in self.SUPPORTED_TARGETS:
            logger.error("Debugging currently not supported for %s", self._target)
            logger.info(
                "Scargo currently supports debug for %s", self.SUPPORTED_TARGETS
            )
            sys.exit(1)

        self._bin_path = bin_path or config.project_root / self._target.get_bin_path(
            config.project.name.lower()
        )
        if not self._bin_path.exists():
            logger.error("Binary %s does not exist", self._bin_path)
            logger.info(
                "Did you run scargo build --profile Debug --target %s?", self._target.id
            )
            sys.exit(1)

    def run_debugger(self) -> None:
        """Run debugger for target"""
        if self._target.id == ScargoTarget.x86:
            self._debug_x86()
        elif self._target.id == ScargoTarget.stm32:
            self._debug_stm32()
        elif self._target.id == ScargoTarget.esp32:
            self._debug_esp32()
        elif self._target.id == ScargoTarget.atsam:
            self._debug_atsam()

    def _debug_x86(self) -> None:
        subprocess.run(["gdb", self._bin_path], check=False)

    def _debug_embedded(self, openocd_args: List[str], gdb_bin: str) -> None:
        openocd_path = find_program_path("openocd")
        openocd_call = ["sudo"] + [openocd_path] + openocd_args

        openocd = subprocess.Popen(openocd_call)  # pylint: disable=consider-using-with
        # Wait for openocd to start
        sleep(1)
        try:
            subprocess.run(
                [
                    gdb_bin,
                    self._bin_path,
                    EMEDDED_GDB_SETTINGS,
                ],
                check=True,
            )
        finally:
            if openocd is not None:
                if platform.system() == "Windows":
                    openocd.terminate()
                else:
                    os.system("sudo pkill -9 -P " + str(openocd.pid))

    def _debug_stm32(self) -> None:
        stm32_config = self._config.get_stm32_config()
        if not stm32_config.chip:
            logger.warning("Chip label not defined in toml. Default to STM32L496AG")
            stm32_config.chip = CHIP_DEFAULTS.get("stm32", "")
        chip_script = f"target/{stm32_config.chip[:7].lower()}x.cfg"
        if not Path("/usr/share/openocd/scripts", chip_script).exists():
            chip_script = f"target/{stm32_config.chip[:7].lower()}.cfg"
        openocd_args = ["-f", ".devcontainer/openocd-script.cfg"]
        self._debug_embedded(openocd_args, "gdb-multiarch")

    def _debug_esp32(self) -> None:
        openocd_args = [
            "-f",
            "interface/ftdi/esp32_devkitj_v1.cfg",
            "-f",
            "board/esp-wroom-32.cfg",
        ]
        self._debug_embedded(openocd_args, "xtensa-esp32-elf-gdb")

    def _debug_atsam(self) -> None:
        openocd_args = ["-f", ".devcontainer/openocd-script.cfg"]
        self._debug_embedded(openocd_args, "gdb-multiarch")


def scargo_debug(bin_path: Optional[Path], target: Optional[ScargoTarget]) -> None:
    config = prepare_config()
    debug = _ScargoDebug(config, bin_path, target)
    debug.run_debugger()



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Create new project"""
import re
import subprocess
import sys
from pathlib import Path
from typing import Dict, List, Optional

from scargo import __version__
from scargo.config import CHIP_DEFAULTS, TARGETS, ScargoTarget, Target
from scargo.config_utils import get_scargo_config_or_exit
from scargo.file_generators.cpp_gen import generate_cpp
from scargo.file_generators.toml_gen import generate_toml
from scargo.global_values import SCARGO_DEFAULT_CONFIG_FILE, SCARGO_DOCKER_ENV
from scargo.logger import get_logger
from scargo.target_helpers.atsam_helper import create_atsam_config
from scargo.target_helpers.esp32_helper import create_esp32_config
from scargo.target_helpers.stm32_helper import create_stm32_config

logger = get_logger()


def get_chip_target(chip_label: str) -> Optional[str]:
    for target in TARGETS:
        if chip_label.startswith(target):
            return target
    return None


def process_chips(chips: List[str], targets: List[ScargoTarget]) -> Dict[str, str]:
    targets_chips = {}
    unsupported_chips = []

    for chip_label in chips:
        chip_label = chip_label.lower()
        chip_target = get_chip_target(chip_label)
        if chip_target is not None:
            targets_chips[chip_target] = chip_label
        else:
            unsupported_chips.append(chip_label)

    if unsupported_chips:
        logger.error(f"Unsupported chips: {unsupported_chips}")
        sys.exit(1)

    for target in targets:
        if target.value not in targets_chips:
            targets_chips[target.value] = CHIP_DEFAULTS[target.value]

    return targets_chips


def scargo_new(
    name: str,
    bin_name: Optional[str],
    lib_name: Optional[str],
    targets: List[ScargoTarget],
    create_docker: bool,
    git: bool,
    chip: List[str],
) -> None:
    """
    Create new project

    :param str name: name of project
    :param Optional[str] bin_name: name of bin file
    :param Optional[str] lib_name: name of lib file
    :param List[Target] targets: target types for a project
    :param bool create_docker: initialize docker environment
    :param bool git: initialize git repository
    :param List[str] chip: list of chips for targets
    :return: None
    :raises FileExistsError: if project with provided name exist
    """
    if not re.match(r"[a-zA-Z][\w-]*$", name):
        logger.error(
            "Name must consist of letters, digits, dash and undescore only,"
            " and the first character must be a letter"
        )
        sys.exit(1)

    targets_chips: Dict[str, str] = process_chips(chip, targets)

    targets.extend([ScargoTarget(key) for key in targets_chips if key not in targets])
    if len(targets) == 0:
        targets = [ScargoTarget.x86]

    # If neither binary target nor library target is specified then create a
    # binary target named same as the project name.
    if not bin_name and not lib_name:
        bin_name = name  # One item tuple.

    project_dir = Path(name)
    try:
        project_dir.mkdir()
    except FileExistsError:
        logger.error("Provided project name: %s already exist.", name)
        sys.exit(1)

    build_env = get_build_env(create_docker)
    targets_ids = ", ".join([f'"{t.name}"' for t in targets])
    if len(targets_ids) > 1:
        targets_ids = f"[{targets_ids}]"

    toml_path = project_dir / SCARGO_DEFAULT_CONFIG_FILE
    generate_toml(
        toml_path,
        project_name=name,
        targets_ids=targets_ids,
        target=[Target.get_target_by_id(t.value) for t in targets],
        build_env=build_env,
        version=__version__,
        docker_image_tag=f"{name.lower()}-dev:1.0",
        lib_name=lib_name,
        bin_name=bin_name,
        atsam_config=create_atsam_config(targets_chips.get("atsam")),
        esp32_config=create_esp32_config(targets_chips.get("esp32")),
        stm32_config=create_stm32_config(targets_chips.get("stm32")),
    )

    config = get_scargo_config_or_exit(toml_path)
    generate_cpp(config)

    test_dir = project_dir / "tests"
    Path(test_dir, "mocks").mkdir(parents=True)
    Path(test_dir, "ut").mkdir(parents=True)
    Path(test_dir, "it").mkdir(parents=True)

    if git:
        subprocess.check_call("git init -q", shell=True, cwd=project_dir)
        logger.info("Initialized git repo")


def get_build_env(create_docker: bool) -> str:
    """
    Get build env
    :param bool create_docker: if create docker
    :return: build env
    """
    if create_docker:
        build_env = f"{SCARGO_DOCKER_ENV}"
    else:
        build_env = "native"
    return build_env



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Build project bin/lib exec file"""
import subprocess
import sys
from pathlib import Path
from typing import List, Optional

from scargo.config import Config, ScargoTarget, Target
from scargo.config_utils import get_target_or_default, prepare_config
from scargo.file_generators.conan_gen import conan_add_default_profile_if_missing
from scargo.logger import get_logger
from scargo.utils.conan_utils import conan_add_remote, conan_source

logger = get_logger()


def scargo_build(
    profile: str, target: Optional[ScargoTarget], all_targets: bool = False
) -> None:
    """
    Build project exec file.

    :param str profile: Profile
    :param target: Target to build
    :param bool all_targets: Build all targets
    :return: None
    """
    config = prepare_config()
    project_dir = config.project_root

    if not Path(project_dir, "CMakeLists.txt").exists():
        logger.error("File `CMakeLists.txt` does not exist.")
        logger.info("Did you run `scargo update`?")
        sys.exit(1)

    conan_add_default_profile_if_missing()
    conan_add_remote(project_dir, config)
    conan_source(project_dir)

    if not all_targets:
        _scargo_build_targets(config, profile, [get_target_or_default(config, target)])
    else:
        _scargo_build_targets(config, profile, config.project.target)


def _scargo_build_targets(config: Config, profile: str, targets: List[Target]) -> None:
    """
    Build project exec file.

    :param str profile: Profile for which to build
    :param list targets: Targets to build
    :return: None
    """
    project_dir = config.project_root
    for build_target in targets:
        logger.info("Building %s target", build_target.id)

        build_dir = Path(project_dir, build_target.get_profile_build_dir(profile))
        build_dir.mkdir(parents=True, exist_ok=True)
        profile_name = build_target.get_conan_profile_name(profile)

        try:
            subprocess.run(
                [
                    "conan",
                    "install",
                    ".",
                    "-pr",
                    f"./config/conan/profiles/{profile_name}",
                    "-of",
                    build_dir,
                    "-b",
                    "missing",
                ],
                cwd=project_dir,
                check=True,
            )
            subprocess.run(
                [
                    "conan",
                    "build",
                    ".",
                    "-pr",
                    f"./config/conan/profiles/{profile_name}",
                    "-of",
                    build_dir,
                ],
                cwd=project_dir,
                check=True,
            )

            logger.info("Copying artifacts...")
            # This is a workaround so that different profiles can work together with conan
            # Conan always calls CMake with '
            subprocess.run(
                f"cp -r -l -f {build_dir}/build/{config.profiles[profile].cmake_build_type}/* .",
                cwd=build_dir,
                shell=True,
                check=True,
            )
            logger.info("Artifacts copied")

        except subprocess.CalledProcessError:
            logger.error("Scargo build target %s failed", build_target.id)
            sys.exit(1)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Clean project from unnecessary files"""
import shutil
from pathlib import Path
from typing import Optional

from scargo.config_utils import get_scargo_config_or_exit
from scargo.logger import get_logger

logger = get_logger()


def _case_insensitive_find_dir(source_dir: Path, dirname: str) -> Optional[Path]:
    if source_dir and source_dir.exists():
        for child in source_dir.iterdir():
            if child.name.lower() == dirname.lower():
                return child
    return None


EXCLUDE_FROM_CLEAN = [".cmake_fetch_cache"]


def handle_item_deletion(item: Path) -> None:
    """Handle the deletion of a file or directory."""
    try:
        if item.is_dir():
            shutil.rmtree(item)
        elif item.is_file():
            item.unlink()
    except (OSError, FileNotFoundError) as e:
        logger.error(f"Error removing {item}: {e}")


def process_directory(directory: Path) -> None:
    """Process each item in the given directory."""
    for item in directory.iterdir():
        if item.name in EXCLUDE_FROM_CLEAN:
            logger.info(f"Skipping clean of: {item}")
            continue
        handle_item_deletion(item)


def scargo_clean() -> None:
    """Clean project dir from unnecessary files"""

    config = get_scargo_config_or_exit()
    project_path = Path(config.project_root)
    test_dir = _case_insensitive_find_dir(project_path, "test")
    source_directories = [project_path, test_dir]

    for source_dir in source_directories:
        if source_dir:
            build_dir = _case_insensitive_find_dir(source_dir, "build")
            if build_dir and build_dir.exists():
                process_directory(build_dir)
                logger.info(f"Cleaned build directory: {build_dir}")
